from __future__ import absolute_import
from .technique import PositionInvariantTechnique
import cv2

class equalizeHistogramAugmentationTechnique(PositionInvariantTechnique):


    def __init__(self,parameters):
        PositionInvariantTechnique.__init__(self, parameters)

    def apply(self, image):
        if len(image.shape) == 3:
            b, g, r = cv2.split(image)
            red = cv2.equalizeHist(r)
            green = cv2.equalizeHist(g)
            blue = cv2.equalizeHist(b)
            return cv2.merge((blue, green, red))
        else:
            return cv2.equalizeHist(image)

# technique = equalizeHistogramAugmentationTechnique()
# image = cv2.imread("LPR1.jpg")
# cv2.imshow("resized",technique.apply(image))
# cv2.waitKey(0)


import imutils
import cv2

class AspectAwarePreprocessor:

    def __init__(self,width,height,inter=cv2.INTER_AREA):
        self.width = width
        self.height = height
        self.inter = inter

    def preprocess(self,image):
        (h, w) = image.shape[:2]
        dW = 0
        dH = 0
        if w < h:
            image = imutils.resize(image, width=self.width,
                                   inter=self.inter)
            dH = int((image.shape[0] - self.height) / 2.0)
        else:
            image = imutils.resize(image, height=self.height,
                               inter=self.inter)
            dW = int((image.shape[1] - self.width) / 2.0)

        (h, w) = image.shape[:2]
        image = image[dH:h - dH, dW:w - dW]

        return cv2.resize(image, (self.width, self.height),
                          interpolation=self.inter)


from __future__ import absolute_import
from .technique import PositionVariantTechnique
import cv2

class cropSizeAugmentationTechnique(PositionVariantTechnique):

    # percentage is a value between 0 and 1
    # startFrom indicates the starting point of the cropping,
    # the possible values are TOPLEFT, TOPRIGHT, BOTTOMLEFT,
    # BOTTOMRIGHT, and CENTER
    def __init__(self,parameters):
        PositionVariantTechnique.__init__(self, parameters)
        if 'x' in list(parameters.keys()):
            self.x = int(parameters["x"])
        else:
            raise NameError("Provide x value")
        if 'y' in list(parameters.keys()):
            self.y = int(parameters["y"])
        else:
            raise NameError("Provide y value")
        if 'width' in list(parameters.keys()):
            self.width = int(parameters["width"])
        else:
            raise NameError("Provide width value")
        if 'height' in list(parameters.keys()):
            self.height = int(parameters["height"])
        else:
            raise NameError("Provide height value")


    def apply(self, image):
        crop = image[self.y:self.y+self.height, self.x:self.x+self.width]
        return crop


# Example
# technique = cropAugmentationTechnique(0.5,'TOPRIGHT')
# image = cv2.imread("LPR1.jpg")
# cv2.imshow("original",image)
# cv2.imshow("new",technique.apply(image))
# cv2.waitKey(0)


from __future__ import absolute_import
from .transformer import Transformer
from .detection import detectBox
from ..techniques.technique import PositionVariantTechnique



class TransformerForImageLocalization(Transformer):

    def __init__(self,technique,dictLabels=None):
        Transformer.__init__(self,technique,dictLabels)


    def transform(self, image,box):
        newImage = self.technique.apply(image)
        if (isinstance(self.technique, PositionVariantTechnique)):
            newBox = detectBox(image, box, self.technique)
            newBox = (self.transformLabel(newBox[0]), newBox[1])
        else:
            newBox = (self.transformLabel(box[0]), box[1])

        return [newImage,newBox]





from __future__ import absolute_import
from builtins import str
from builtins import object
from .iaugmentor import IAugmentor
from imutils import paths
import os
import cv2
from sklearn.externals.joblib import Parallel, delayed
import xml.etree.ElementTree as ET
from .utils import prettify



def generateXML(filename,outputPath,w,h,d,boxes):
    top = ET.Element('annotation')
    childFolder = ET.SubElement(top, 'folder')
    childFolder.text = 'images'
    childFilename = ET.SubElement(top, 'filename')
    childFilename.text = filename[0:filename.rfind(".")]
    childPath = ET.SubElement(top, 'path')
    childPath.text = outputPath + "/" + filename
    childSource = ET.SubElement(top, 'source')
    childDatabase = ET.SubElement(childSource, 'database')
    childDatabase.text = 'Unknown'
    childSize = ET.SubElement(top, 'size')
    childWidth = ET.SubElement(childSize, 'width')
    childWidth.text = str(w)
    childHeight = ET.SubElement(childSize, 'height')
    childHeight.text = str(h)
    childDepth = ET.SubElement(childSize, 'depth')
    childDepth.text = str(d)
    childSegmented = ET.SubElement(top, 'segmented')
    childSegmented.text = str(0)
    for box in boxes:
        (category, (x,y,wb,hb)) = box
        childObject = ET.SubElement(top, 'object')
        childName = ET.SubElement(childObject, 'name')
        childName.text = category
        childPose = ET.SubElement(childObject, 'pose')
        childPose.text = 'Unspecified'
        childTruncated = ET.SubElement(childObject, 'truncated')
        childTruncated.text = '0'
        childDifficult = ET.SubElement(childObject, 'difficult')
        childDifficult.text = '0'
        childBndBox = ET.SubElement(childObject, 'bndbox')
        childXmin = ET.SubElement(childBndBox, 'xmin')
        childXmin.text = str(x)
        childYmin = ET.SubElement(childBndBox, 'ymin')
        childYmin.text = str(y)
        childXmax = ET.SubElement(childBndBox, 'xmax')
        childXmax.text = str(x+wb)
        childYmax = ET.SubElement(childBndBox, 'ymax')
        childYmax.text = str(y+hb)
    return prettify(top)

#
def readAndGenerateImage(outputPath, transformers, i_and_imagePath):
    (i, imagePath) = i_and_imagePath
    image = cv2.imread(imagePath)
    name = imagePath.split(os.path.sep)[-1]
    labelPath = '/'.join(imagePath.split(os.path.sep)[:-1]) + "/"+name[0:name.rfind(".")] + ".xml"
    tree = ET.parse(labelPath)
    root = tree.getroot()
    objects = root.findall('object')
    if(len(objects)!=1):
        raise Exception("Since this is a localization problem, the xml should only contain one object")
    object = objects[0]
    category = object.find('name').text
    bndbox = object.find('bndbox')
    x  = int(bndbox.find('xmin').text)
    y = int(bndbox.find('ymin').text)
    h = int(bndbox.find('ymax').text)-y
    w = int(bndbox.find('xmax').text) - x

    for (j, transformer) in enumerate(transformers):
        (newimage, box) = transformer.transform(image, (category, (x, y, w, h)))
        if box is not None:
            cv2.imwrite(outputPath + "/" + str(i) + "_" + str(j) + "_" + name,
                        newimage)
            (hI,wI) =image.shape[:2]
            if(len(image.shape)==3):
                d = 3
            else:
                d=1
            file = open(outputPath + "/" + str(i) + "_" + str(j) + "_" + name[0:name.rfind(".")]+".xml", "w")
            file.write(generateXML(str(i) + "_" + str(j) + "_" + name,outputPath,wI,hI,d,[box]))
            file.close()


# # This class serves to generate images for a localization
# # problem where all the images in the given folder, and the labels
# # are given in the same folder with the same name and using the PASCAL VOC format.
# # Example:
# # - Folder
# # |- image1.jpg
# # |- image1.xml
# # |- image2.jpg
# # |- image2.xml
# # |- ...
# #
#
class PascalVOCLinearLocalizationAugmentor(IAugmentor):

    def __init__(self,inputPath,parameters):
        IAugmentor.__init__(self)
        self.inputPath = inputPath
        # output path represents the folder where the images will be stored
        if parameters["outputPath"]:
            self.outputPath = parameters["outputPath"]
        else:
            raise ValueError("You should provide an output path in the parameters")


    def readImagesAndAnnotations(self):
        self.imagePaths = list(paths.list_files(self.inputPath,validExts=(".jpg", ".jpeg", ".png", ".bmp",".tiff",".tif")))
        self.labelPaths = list(paths.list_files(self.inputPath,validExts=(".xml")))
        if (len(self.imagePaths) != len(self.labelPaths)):
            raise Exception("The number of images is different to the number of annotations")

    def applyAugmentation(self):
        self.readImagesAndAnnotations()
        #[readAndGenerateImage(self.outputPath, self.generators, x) for x in enumerate(self.imagePaths)]
        Parallel(n_jobs=-1)(delayed(readAndGenerateImage)(self.outputPath,self.transformers,x) for x in enumerate(self.imagePaths))




from __future__ import absolute_import
from .augmentorsList import Augmentors

def createAugmentor(problem,annotationmode,outputmode,generationmode,inputPath,parameters):
    Augmentor = Augmentors[problem][annotationmode][outputmode][generationmode]
    if Augmentor is None:
        raise ValueError("These parameters are not currently supported")

    return Augmentor(inputPath,parameters)



from __future__ import absolute_import
from .technique import PositionInvariantTechnique
import cv2
import numpy as np

class blurringAugmentationTechnique(PositionInvariantTechnique):

    # Valid values for ksize are 3,5,7,9, and 11
    def __init__(self,parameters):
        PositionInvariantTechnique.__init__(self, parameters)
        if 'ksize' in list(parameters.keys()):
            self.ksize = int(parameters["ksize"])
        else:
            self.ksize = 3
        if (not (self.ksize in [3,5,7,9,11])):
            raise NameError("Invalid value for ksize")


    def apply(self, image):
        blurred = cv2.blur(image, (self.ksize,self.ksize))
        return blurred


# technique = blurringAugmentationTechnique(5)
# image = cv2.imread("LPR1.jpg")
# cv2.imshow("resized",technique.apply(image))
# cv2.waitKey(0)


from __future__ import absolute_import
from keras.optimizers import SGD
import argparse
from .utils.conf import Conf
from .augmentors.augmentorFactory import createAugmentor
from .transformers.transformerFactory import transformerGenerator
from .techniques.techniqueFactory import createTechnique
from .utils.minivgg import MiniVGGNet

ap = argparse.ArgumentParser()
ap.add_argument("-c", "--conf", required=True, help="path to configuration file")
args = vars(ap.parse_args())

conf = Conf(args["conf"])
# First, we read the parameters
problem = conf["problem"]
annotationMode = conf["annotation_mode"]
outputMode = conf["output_mode"]

if not(outputMode=="keras"):
    exit()

generationMode = conf["generation_mode"]
inputPath = conf["input_path"]
parameters = conf["parameters"]
augmentationTechniques = conf["augmentation_techniques"]

# Second, we create the augmentor
augmentor = createAugmentor(problem,annotationMode,outputMode,generationMode,inputPath,
                            parameters)

# Third, we load the techniques and add them to the augmentor
techniques = [createTechnique(technique,param) for (technique,param) in
              augmentationTechniques]

transformer = transformerGenerator(problem)

for technique in techniques:
    augmentor.addTransformer(transformer(technique))


# Definition of the model
opt = SGD(lr=0.05)
model = MiniVGGNet.build(width=parameters["width"], height=parameters["height"], depth=3)
model.compile(loss="categorical_crossentropy", optimizer=opt,metrics=["accuracy"])

# Training the model
H = model.fit_generator(augmentor.applyAugmentation(),
                         steps_per_epoch= parameters["batchSize"],
                         epochs=100, verbose=1)



import xml.etree.ElementTree as ET
from xml.dom import minidom


def prettify(elem):
    """Return a pretty-printed XML string for the Element.
    """
    rough_string = ET.tostring(elem, 'utf-8')
    reparsed = minidom.parseString(rough_string)
    return reparsed.toprettyxml(indent="  ")


from __future__ import absolute_import
from .transformerForImageClassification import TransformerForImageClassification
from .transformerForImageDetection import TransformerForImageDetection
from .transformerForImageInstanceSegmentation import TransformerForImageInstanceSegmentation
from .transformerForImageLocalization import TransformerForImageLocalization
from .transformerForImageSegmentation import TransformerForImageSegmentation
from .transformerForImageStackClassification import TransformerForImageStackClassification


Transformers = {
    "classification" : TransformerForImageClassification,
    "localization" : TransformerForImageLocalization,
    "detection" : TransformerForImageDetection,
    "semantic_segmentation" : TransformerForImageSegmentation,
    "instance_segmentation" : TransformerForImageInstanceSegmentation,
    "stackclassification" : TransformerForImageStackClassification,
    
}